#!/usr/bin/python
# -*- coding: utf-8 -*-

from scannedonlyd import ScannedOnly, SCANNEDONLY, Options

class Scanner(object):
    def __init__(self, sock):
        self.rconn = sock.makefile('r', 1)
        self.conn = sock
        self.__id = 0
        #print ('>>', 'nIDSESSION\n')
        self.conn.sendall('nIDSESSION\n')

    def __del__(self):
        #print ('>>', 'nEND\n')
        self.conn.sendall('nEND\n')

    def read_fd(self):
        return self.rconn.fileno()

    def write_fd(self):
        return self.conn.fileno()

    def produce(self, path):
        self.__id += 1
        #print ('>>', 'nSCAN %s\n' % path)
        self.conn.sendall('nSCAN %s\n' % path)
        return path

    def consume(self, path):
        line = self.rconn.readline()[:-1]     # chomp '\n'
        #print ('<<', '%s' % line)
        session_id, result = line.split(': ', 1)
        result, result_code = result.rsplit(None, 1)
        if ': ' in result:
            _, result = result.rsplit(': ', 1)
        #print (session_id, result, result_code)
        if result_code == 'OK':
            return SCANNEDONLY.CLEAN, path, result
        if result_code == 'FOUND':
            return SCANNEDONLY.INFECTED, path, result
        if result_code == 'FAILED':
            return SCANNEDONLY.FAILED, path, result

class Clamd(object):
    def __init__(self, opts):
        self.opts = opts

    def setup(self):
        pass

    def scanners(self):
        import socket
        if self.opts.clamd_socket:
            socktype = (socket.AF_UNIX, socket.SOCK_DGRAM)
            sockaddr = self.opts.clamd_socket
        else:
            socktype = (socket.AF_INET, socket.SOCK_STREAM)
            sockaddr = (self.opts.clamd_ip, self.opts.clamd_port)
        sock = socket(*socktype)
        for i in range(self.opts.maxthreads):
            sock.connect(sockaddr)
            yield Scanner(sock)
            sock = socket(*socktype)

if __name__ == '__main__':
    from optparse import Option
    ip = Option('-I', '--clamd_ip', help="IP of Clam daemon "
                "(default: %default)", default="127.0.0.1")
    ports = Option('-P', '--clamd_port', type="int",
                   help="TCP port number Clamd daemon listens on ")
    unix = Option('-U', '--clamd_socket',
                  help="Path to the (Unix) socket Clamd daemon listens on")
    opts, args = Options(ip, ports, unix)
    ScannedOnly(opts)(Clamd(opts))


