/* 
 * scannedonly_prescan  
 *
 * Copyright 2007,2008 (C) Olivier Sessink
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#define _GNU_SOURCE
#include <sys/types.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>
#include <string.h>

#include "config.h"

extern int optind;

#define SENDBUFFERSIZE 1450

typedef struct {
	int socket;
	int domain_socket;
	char *socketname;
	int portnum;
	char *scanhost;
	int bytes;
	int files;
	
	int sizecounter;
	int numcounter;
	int timecounter;
	char gsendbuffer[SENDBUFFERSIZE + 1];
} Tscannedonly;

Tscannedonly *so;

#ifndef SUN_LEN
	#define SUN_LEN(su) (sizeof (*(su)) - sizeof ((su)->sun_path) + strlen((su)->sun_path))
#endif

#ifndef HAVE_STRNDUP
#include <sys/ddi.h>
char *strndup( const char *s, size_t n ) {
	size_t nAvail;
	char *p;
	if ( !s )
		return 0;
	nAvail = min(strlen(s) + 1, n + 1 );
	p = malloc( nAvail );
	memcpy( p, s, nAvail );
	p[nAvail - 1] = '\0';
	return p;
}
#endif

static char *name_w_ending_slash(const char *name) {
	int len;
	
	len = strlen(name);
	if (name[len-1] == '/') {
		return strdup(name);
	} else {
		char *tmp;
		tmp = malloc(len+2);
		strncpy(tmp, name, len+1);
		return strncat(tmp, "/", 1);
	}
}

static char *path_plus_name(const char *basename, const char *filename) {
	int baselen = strlen(basename);
	int filelen = strlen(filename);
	
	return strncat(strncpy(malloc(baselen+filelen+1),basename,baselen+1),filename,filelen+1);
}

static char *cachefile_name(const char *shortname, const char *basename) {
	char *cachefile;
	int rlen = strlen(shortname);
	int blen = strlen(basename);
	/*printf( "cachefile_name shortname=%s basename=%s\n", shortname, basename );*/
	cachefile = (char *)malloc(rlen + blen + 9 + 1);
	cachefile = strncat(strncat(strncpy(cachefile, basename, blen+1),".scanned:", 9+1), shortname, rlen+1);
	return cachefile;
}

static char *construct_full_path(const char *relpath, int ending_slash) {
	char *fullpath, tmp[1024];
	const char *iter;
	int tmplen,rellen,itlen;
	
	getcwd(tmp, 1024);
	tmplen = strlen(tmp);
	rellen = strlen(relpath);
	fullpath = malloc(tmplen+rellen+4);
	fullpath = strncpy(fullpath, tmp, tmplen+1);
	if (fullpath[tmplen-1] != '/') {
		fullpath = strncat(fullpath, "/", 2);
	}
	iter = relpath;
	while (*iter == '.' && *(iter+1) == '/') 
		iter += 2; 
	fullpath = strncat(fullpath, iter, rellen+1);
	/* iter can be empty */
	itlen = strlen(iter);
	if (ending_slash && itlen > 0 && iter[itlen-1] != '/') {
		fullpath = strncat(fullpath, "/", 2);
	}
	return fullpath;
}


static int connect_to_scanner(void) {
	
	if (so->domain_socket) {
		struct sockaddr_un saun;
		printf( "socket=%s\n",so->socketname);
		if ((so->socket = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0) {
			printf( "failed to create socket\n");
			return -1;
		}
		saun.sun_family = AF_UNIX;
		strncpy(saun.sun_path, so->socketname,strlen(so->socketname)+1);
		if (connect(so->socket, (struct sockaddr *)&saun, SUN_LEN(&saun)) < 0) {
			printf ( "failed to connect to socket\n");
			return -1;
		}
		printf( "bound %s to socket %d\n", saun.sun_path, so->socket);
		
	} else {
		struct sockaddr_in sain;
		struct hostent *he;
		printf( "scanhost=%s, port=%d\n",so->scanhost,so->portnum);
		if ((he=gethostbyname(so->scanhost))==NULL) {
			printf( "failed to resolve hostname\n");
		}
		if ((so->socket = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			printf( "failed to create socket\n");
			return -1;
		}
		memset(&sain, 0, sizeof(sain));
		sain.sin_family = AF_INET;
		sain.sin_port = htons(so->portnum);
		sain.sin_addr = *((struct in_addr *)he->h_addr);

		if (connect(so->socket, (struct sockaddr *)&sain, sizeof(sain)) < 0) {
			printf( "failed to connect to socket\n");
			return -1;
		}
		printf( "bound port %d to socket %d\n", so->portnum, so->socket);
	}

	{
		int sndsize;
		socklen_t size=sizeof(int);
		getsockopt(so->socket, SOL_SOCKET, SO_RCVBUF, (char *)&sndsize, &size);
		printf( "current socket buffer size=%d\n",sndsize);
		sndsize = 262144;
		if (setsockopt(so->socket, SOL_SOCKET, SO_RCVBUF, (char *)&sndsize,(int)sizeof(sndsize))!=0) {
			printf( "error setting socket buffer %s (%d)\n", strerror(errno), errno);
		}
	}


	if (fcntl(so->socket, F_SETFL, O_NDELAY) < 0) {
		printf( "error setting socket NONBLOCKING %s (%d)\n", strerror(errno), errno);
	}
	
	return 0;
}

static void flush_sendbuffer(void)
{
	int ret, len, loop = 10;
	if (so->gsendbuffer[0] == '\0') {
		return;
	}

	do {
		loop--;
		len = strlen(so->gsendbuffer);
		ret = send(so->socket, so->gsendbuffer, len, MSG_DONTWAIT);
		if (ret == len) {
			so->gsendbuffer[0] = '\0';
			break;
		}
		if (ret == -1) {
			if (errno == ECONNREFUSED || errno == ENOTCONN
			    || errno == ECONNRESET) {
				if (connect_to_scanner() == -1)
					break;	/* connecting fails, abort */
				/* try again */
			} else if (errno != EINTR) {
				/* on EINTR we just try again, all remaining
				   other errors we log the error
				   and try again ONCE */
				loop = 1;
			}
		} else {
			/* --> partial write: Resend all filenames that were
			   not or not completely written. a partial filename
			   written means the filename will not arrive correctly,
			   so resend it completely */
			int pos = 0;
			while (pos < len) {
				char *tmp = strchr(so->gsendbuffer+pos, '\n');
				if (tmp && tmp - so->gsendbuffer < ret)
					pos = tmp - so->gsendbuffer + 1;
				else
					break;
			}
			memmove(so->gsendbuffer, so->gsendbuffer + pos,
				SENDBUFFERSIZE - ret);
			/* now try again */
		}
	} while (loop > 0);

	if (so->gsendbuffer[0] != '\0') {
		so->gsendbuffer[0] = '\0';
	}
}

static void notify_scanner(const char *scanfile, int size)
{
	char *tmp;
	int tmplen, gsendlen, free_tmp=0;
	
	printf("notify_scanner: %s needs scanning\n", scanfile);
	
	if (scanfile[0] != '/') {
		tmp = construct_full_path(scanfile, 0);
		free_tmp=1;
	} else {
		tmp = (char *)scanfile;
	}
	tmplen = strlen(tmp);
	gsendlen = strlen(so->gsendbuffer);
	if (gsendlen + tmplen >= SENDBUFFERSIZE) {
		flush_sendbuffer();
	}
	strncat(so->gsendbuffer, tmp, SENDBUFFERSIZE + 1);
	strncat(so->gsendbuffer, "\n", SENDBUFFERSIZE + 1);
	if (free_tmp)
		free(tmp);

	/* don't overflow the scanner with requests */
	so->sizecounter += size;
	so->numcounter++;
	if (so->sizecounter >= so->bytes || so->numcounter >= so->files) {
		int sleeptime = time(NULL) - so->timecounter + 60;
		
		if (sleeptime > 0) {
			printf("sent %d files (%d Mb): sleeping for %d secons\n",so->numcounter,so->sizecounter/(1024*1024),sleeptime);
			sleep(sleeptime);
		} 
		so->sizecounter -= so->bytes;
		if (so->sizecounter < 0) so->sizecounter = 0; 
		so->numcounter -= so->files;
		if (so->numcounter < 0) so->numcounter = 0;
		so->timecounter = time(NULL);
	}
}

static void rundir(char *dir);
static void rundirentry(char *basedir, char *entry) {
	char *name;
	struct stat sbuf1, sbuf2;
	name = path_plus_name(basedir, entry);
	if (lstat(name, &sbuf1)!=-1) {
		if (S_ISDIR(sbuf1.st_mode)) {
			char *dirname;
			dirname = name_w_ending_slash(name);
			rundir(dirname);
			free(dirname);
		} else if (S_ISREG(sbuf1.st_mode) && sbuf1.st_size!= 0){
			char * cachedname;
			int ret;
			cachedname = cachefile_name(entry, basedir);
			ret = lstat(cachedname,&sbuf2);
			if (ret == -1) {
				notify_scanner(name,sbuf1.st_size);
			} else {
				if (sbuf1.st_ctime > sbuf2.st_ctime) {
					notify_scanner(name,sbuf1.st_size);
				}
			}
			free(cachedname);
		}
	}
	free(name);
}

static void rundir(char *dir) {
	DIR *dirp;
	struct dirent *dirt;
	dirp = opendir(dir);
	if (!dirp)
		return;
	printf("reading directory %s\n",dir);
	while ((dirt = readdir(dirp)) != NULL) {
		if (strcmp(dirt->d_name, ".")==0 
				|| strcmp(dirt->d_name, "..")==0
				|| strncmp(dirt->d_name, ".scanned:", 9)==0
				|| strncmp(dirt->d_name, ".virus:", 7)==0
				|| strncmp(dirt->d_name, ".failed:", 8)==0
				) {
			continue; 	
		} 
		rundirentry(dir, dirt->d_name);
	}
	closedir(dirp);
}

static void print_usage() {
	printf("\nUsage: scannedonly_prescan <options> /path /path /path\n\n");
	printf("-s --socket   : specify socket name to send to (default /var/lib/scannedonly/scan)\n");
	printf("-p --portnum  : specify the UDP port to send to (default 2020)\n");
	printf("-n --scanhost : specify hostname of the scanhost (default localhost)\n");
	printf("-m --megabytes: specify the max. number of megabytes per minute to send (default 50)\n");
	printf("-f --files    : specify the max. number of files per minute to send (default 1000)\n");
	printf("\nuse either -s OR -p but not both. If nothing is specified scannedonly_prescan will\n");
	printf("send to the domain socket\n\n");
}

int main(int argc, char **argv) {
	int i, c=0;
	
	so = malloc(sizeof(Tscannedonly));
	so->socket=0;
	so->domain_socket = 1;
	so->socketname = NULL;
	so->portnum = 2020;
	so->scanhost = "localhost";
	so->bytes = 50 * 1024*1024;
	so->files = 1000;
	so->timecounter = 0;
	so->sizecounter = 0;
	so->numcounter = 0;
	while (c!=-1) {
		int option_index;
		static struct option long_options[] = {
			{"portnum", required_argument, NULL, 'p'},
			{"socket", required_argument, NULL, 's'},
			{"scanhost", required_argument, NULL, 'n'},
			{"megabytes", required_argument, NULL, 'm'},
			{"files", required_argument, NULL, 'f'},
			{"help", no_argument, NULL, 'h'},
			{NULL,0,NULL,0}		
		};

		c = getopt_long(argc,argv, "n:m:f:p:s:h?", long_options, &option_index);
		
		switch (c) {
		case 'p':
			if (so->socketname != NULL) {
				printf("ABORT: Invalid combination of options: cannot set domain socket AND UDP port number\n");
				exit(1);
			}
			so->portnum = atoi(optarg);
			so->domain_socket=0;
		break;
		case 's':
			if (so->domain_socket==0) {
				printf("ABORT: Invalid combination of options: cannot set domain socket AND UDP port number\n");
				exit(1);
			}
			so->socketname = strdup(optarg);
		break;
		case 'n':
			so->scanhost = strdup(optarg);
		break;
		case 'm':
			so->bytes = atoi(optarg) * 1024*1024;
		break;
		case 'f':
			so->files = atoi(optarg);
		break;
		case 'h':
		case '?':
			print_usage();
			exit(3);
		}
	}
	if (so->socketname == NULL && so->domain_socket) {
		so->socketname = "/var/lib/scannedonly/scan";
	}
	connect_to_scanner();
	so->timecounter = time(NULL);
	for (i=optind;i<argc;i++) {
		if (argv[i]) {
			char *dir;
			struct stat sbuf1;
			if (lstat(argv[i], &sbuf1)!=-1) {
				if (S_ISDIR(sbuf1.st_mode)) {
					if (argv[i][0] != '/') {
						dir = construct_full_path(argv[i], 1);
					} else {
						dir = name_w_ending_slash(argv[i]);
					}
					rundir(dir);
					free(dir);
				} else if (S_ISREG(sbuf1.st_mode)) {
					char *entry, *basedir;
					entry = strrchr(argv[i], '/') + 1;
					basedir = strndup(argv[i], strlen(argv[i])-strlen(entry));
					rundirentry(basedir, entry);
					free(basedir);
				}
			} else {
				printf("Warning: cannot stat %s: %s",argv[i],strerror(errno));
			}
		}
	}
	flush_sendbuffer();
	return 0;
}
